import logging
import time
from typing import Any, Dict, List, Optional, Union

from paradex_py.account.account import ParadexAccount
from paradex_py.api.http_client import HttpClient, HttpMethod
from paradex_py.api.models import AccountSummary, AccountSummarySchema, AuthSchema, SystemConfig, SystemConfigSchema
from paradex_py.common.order import Order
from paradex_py.environment import Environment


class ParadexApiClient(HttpClient):
    """ParadexApiClient class to interact with Paradex API.
    Initialized along with Paradex class.

    Args:
        env (Environment): Environment
        logger (logging.Logger, optional): Logger. Defaults to None.

    Examples:
        >>> from paradex_py import Paradex
        >>> from paradex_py.environment import Environment
        >>> paradex = Paradex(env=Environment.TESTNET)
        >>> paradex.ws_client.connect()
    """

    def __init__(
        self,
        env: Environment,
        logger: Optional[logging.Logger] = None,
    ):
        self.env = env
        self.logger = logger or logging.getLogger(__name__)
        super().__init__()
        self.api_url = f"https://api.{self.env}.paradex.trade/v1"

    async def __aexit__(self):
        await self.client.close()

    def load_system_config(self) -> SystemConfig:
        res = self.request(
            url=f"{self.api_url}/system/config",
            http_method=HttpMethod.GET,
        )
        self.logger.info(f"ParadexApiClient: /system/config:{res}")
        config = SystemConfigSchema().load(res)
        self.logger.info(f"ParadexApiClient: SystemConfig:{config}")
        return config

    def init_account(self, account: ParadexAccount):
        self.account = account
        self.onboarding()
        self.auth()

    def onboarding(self):
        headers = self.account.onboarding_headers()
        payload = {"public_key": hex(self.account.l2_public_key)}
        self.post(api_url=self.api_url, path="onboarding", headers=headers, payload=payload)

    def auth(self):
        headers = self.account.auth_headers()
        res = self.post(api_url=self.api_url, path="auth", headers=headers)
        data = AuthSchema().load(res)
        self.auth_timestamp = time.time()
        self.account.set_jwt_token(data.jwt_token)
        self.client.headers.update({"Authorization": f"Bearer {data.jwt_token}"})
        self.logger.info(f"ParadexApiClient: JWT:{data.jwt_token}")

    def _validate_auth(self):
        if self.account is None:
            raise ValueError("ParadexApiClient: Account not found")
        # Refresh JWT if it's older than 4 minutes
        if time.time() - self.auth_timestamp > 4 * 60:
            self.auth(headers=self.account.auth_headers())

    def _get(self, path: str, params: Optional[dict] = None) -> dict:
        return self.get(api_url=self.api_url, path=path, params=params)

    def _get_authorized(self, path: str, params: Optional[dict] = None) -> dict:
        self._validate_auth()
        return self._get(path=path, params=params)

    def _post_authorized(
        self,
        path: str,
        payload: Optional[Union[Dict[str, Any], List[Dict[str, Any]]]] = None,
        params: Optional[dict] = None,
        headers: Optional[dict] = None,
    ) -> dict:
        self._validate_auth()
        return self.post(api_url=self.api_url, path=path, payload=payload, params=params, headers=headers)

    def _delete_authorized(self, path: str) -> dict:
        self._validate_auth()
        return self.delete(api_url=self.api_url, path=path)

    # PRIVATE GET METHODS
    def fetch_orders(self, market: str) -> Optional[List]:
        """Fetch open orders for the account.
            Private call requires authorization.

        Args:
            market (str): instrument symbol. If empty then fetch orders for all markets.
        Returns:
            None: if received invalid response from Paradex API
            List: list of dictionaries, each dict represent an order
        """
        params = {"market": market} if market else {}
        response = self._get_authorized(path="orders", params=params)
        return response.get("results") if response else None

    def fetch_orders_history(self, market: str = "") -> Optional[List]:
        """Fetch history of orders for the account.
            Private call requires authorization.

        Args:
            market (str): instrument symbol. If None or empty then fetch history of orders for all markets.
        Returns:
            None: if received invalid response from Paradex API
            List: list of dictionaries, each dict representing an order
        """
        params = {"market": market} if market else {}
        response = self._get_authorized(path="orders-history", params=params)
        return response.get("results") if response else None

    def fetch_order(self, order_id: str) -> Optional[Dict]:
        """Fetch a state of specific order sent from this account.
            Private call requires authorization.

        Args:
            order_id (str): order's id as assigned by Paradex.
        Returns:
            None: if received invalid response from Paradex API
            Dict: dictionary representing an order
        """
        path: str = f"orders/{order_id}"
        return self._get_authorized(path=path)

    def fetch_order_by_client_id(self, client_id: str) -> Optional[Dict]:
        """Fetch a state of specific order sent from this account.
            Private call requires authorization.

        Args:
            client_id (str): order's client_id as assigned by a trader.
        Returns:
            None: if received invalid response from Paradex API
            Dict: dictionary representing an order
        """
        path: str = f"orders/by_client_id/{client_id}"
        return self._get_authorized(path=path)

    def fetch_fills(self, market: str = "") -> Optional[List]:
        """Fetch history of fills for this account.
            Private call requires authorization.

        Args:
            market (str): instrument symbol. If None or empty then fetch fills for all markets.
        Returns:
            None: if received invalid response from Paradex API
            List: list of dictionaries, each dict representing a fill
        """
        params = {"market": market} if market else {}
        response = self._get_authorized(path="fills", params=params)
        return response.get("results") if response else None

    def fetch_funding_payments(self, market: str = "ALL") -> Optional[List]:
        """Fetch history of funding payments for this account.
            Private call requires authorization.

        Args:
            market (str): instrument symbol. If 'ALL' then fetch funding payments for all markets.
        Returns:
            None: if received invalid response from Paradex API
            List: list of dictionaries, each dict representing a funding payment
        """
        params = {"market": market} if market else {}
        response = self._get_authorized(path="funding/payments", params=params)
        return response.get("results") if response else None

    def fetch_transactions(self) -> Optional[List]:
        """Fetch history of transactions initiated by this account.
            Private call requires authorization.

        Returns:
            None: if received invalid response from Paradex API
            List: list of dictionaries, each dict representing a transaction
        """
        response = self._get_authorized(path="transactions")
        return response.get("results") if response else None

    def fetch_account_summary(self) -> AccountSummary:
        """Fetch current summary for this account.
            Private call requires authorization.

        Returns:
            AccountSummary: objects with fields representing account summary
        """
        res = self._get_authorized(path="account")
        return AccountSummarySchema().load(res)

    def fetch_balances(self) -> Optional[List]:
        """Fetch all coin balances for this account
            Private call requires authorization.

        Returns:
            None: if received invalid response from Paradex API
            List: list of dictionaries, each dict representing a balance in specific coin
        """
        response = self._get_authorized(path="balance")
        return response.get("results") if response else None

    def fetch_positions(self) -> Optional[List]:
        """Fetch all derivatives positions for this account
            Private call requires authorization.

        Returns:
            None: if received invalid response from Paradex API
            List: list of dictionaries, each dict representing a position in specific instrument
        """
        response = self._get_authorized(path="positions")
        return response.get("results") if response else None

    # PUBLIC GET METHODS
    def fetch_markets(self) -> Optional[List]:
        """Fetch all markets information
            Public call, no authorization required.

        Returns:
            None: if received invalid response from Paradex API
            List: list of dictionaries, each dict representing a market
        """
        response = self._get(path="markets")
        return response.get("results") if response else None

    def fetch_markets_summary(self, market: str) -> Optional[List]:
        """Fetch ticker information for specific market
            Public call, no authorization required.

        Args:
            market (str): instrument symbol. If 'ALL' then fetch tickers for all markets.
        Returns:
            None: if received invalid response from Paradex API
            List: list of dictionaries, each dict representing a ticker for specific market
        """
        response = self._get(path="markets/summary", params={"market": market})
        return response.get("results") if response else None

    def fetch_orderbook(self, market: str) -> dict:
        """Fetch order-book for specific market
            Public call, no authorization required.

        Args:
            market (str): instrument symbol.
        Returns:
            None: if received invalid response from Paradex API
            Dict: dictionary representing a full order book depth for specific market
        """
        return self._get(path=f"orderbook/{market}")

    def fetch_insurance_fund(self) -> Optional[Dict[Any, Any]]:
        """Fetch insurance fund information
            Public call, no authorization required.

        Returns:
            None: if received invalid response from Paradex API
            Dict: dictionary representing a state of Paradex Insurance Fund
        """
        return self._get(path="insurance")

    def fetch_trades(self, market: str) -> Optional[List]:
        """Fetch Paradex exchange trades for specific market
            Public call, no authorization required.

        Args:
            market (str): instrument symbol.
        Returns:
            None: if received invalid response from Paradex API
            Dict: list of dictionaries, dictionary representing a state of Paradex Insurance Fund
        """
        response = self._get(path="trades", params={"market": market})
        return response.get("results") if response else None

    # order helper functions
    def submit_order(self, order: Order) -> Optional[Dict]:
        """Send order to Paradex
            Private call requires authorization.

        Args:
            order (Order): Order object
        Returns:
            None: if received invalid response from Paradex API
            Dict: dictionary representing a response from Paradex API
        """
        response = None
        order.signature = self.account.sign_order(order)
        order_payload = order.dump_to_dict()
        try:
            response = self._post_authorized(path="orders", payload=order_payload)
        except Exception:
            self.logger.exception(f"submit_order payload:{order_payload}")
        return response

    def cancel_order(self, order_id: str) -> None:
        """Cancel open order previously sent to Paradex from this account.
            Private call requires authorization.

        Args:
            order_id (str): Order id as assigned by Paradex.
        Returns:
            None
        """
        self._delete_authorized(path=f"orders/{order_id}")

    def cancel_order_by_client_id(self, client_id: str) -> None:
        """Cancel open order previously sent to Paradex from this account.
            Private call requires authorization.

        Args:
            client_id (str): Order id as assigned by a trader.
        Returns:
            None
        """
        self._delete_authorized(path=f"orders/by_client_id/{client_id}")



from typing import List, Optional

from starknet_py.net.account.account import Account as StarknetAccount
from starknet_py.net.client import Client
from starknet_py.net.models import AddressRepresentation, StarknetChainId
from starknet_py.net.signer import BaseSigner
from starknet_py.net.signer.stark_curve_signer import KeyPair
from starknet_py.utils.typed_data import TypedData as TypedDataDataclass

from .typed_data import TypedData
from .utils import message_signature


class Account(StarknetAccount):
    def __init__(
        self,
        *,
        address: AddressRepresentation,
        client: Client,
        signer: Optional[BaseSigner] = None,
        key_pair: Optional[KeyPair] = None,
        chain: Optional[StarknetChainId] = None,
    ):
        super().__init__(address=address, client=client, signer=signer, key_pair=key_pair, chain=chain)

    def sign_message(self, typed_data: TypedData) -> List[int]:
        typed_data_dataclass = TypedDataDataclass.from_dict(typed_data)
        msg_hash = typed_data_dataclass.message_hash(self.address)
        r, s = message_signature(msg_hash=msg_hash, priv_key=self.signer.key_pair.private_key)
        return [r, s]



import asyncio
import json
import logging
import time
import traceback
from enum import Enum
from typing import Optional

import websockets

from paradex_py.account.account import ParadexAccount
from paradex_py.environment import Environment


class ParadexWebsocketChannel(Enum):
    """ParadexWebsocketChannel Enum class to define the channels for Paradex WS API.

    Attributes:
        FILLS (str): Fills channel
        ACCOUNT (str): Account channel
        MARKETS_SUMMARY (str): Markets summary channel
        ORDERS (str): Orders channel
        ORDER_BOOK (str): Order book channel
        POSITIONS (str): Positions channel
        TRADES (str): Trades channel
        TRADEBUSTS (str): Tradebusts channel
        TRANSACTIONS (str): Transactions channel
        BALANCE_EVENTS (str): Balance events channel
        FUNDING_DATA (str): Funding data channel
    """

    FILLS = "fills.{market}"
    ACCOUNT = "account"
    MARKETS_SUMMARY = "markets_summary"
    ORDERS = "orders.{market}"
    ORDER_BOOK = "order_book.{market}.snapshot@15@100ms"
    POSITIONS = "positions"
    TRADES = "trades.{market}"
    TRADEBUSTS = "tradebusts"
    TRANSACTIONS = "transaction"
    BALANCE_EVENTS = "balance_events"
    FUNDING_DATA = "funding_data.{market}"

    def prefix(self) -> str:
        return self.value.split(".")[0]


class ParadexWebsocketClient:
    """ParadexWebsocketChannel class to interact with Paradex WS API.
    Initialized along with Paradex class.

    Args:
        env (Environment): Environment
        logger (Optional[logging.Logger], optional): Logger. Defaults to None.

    Examples:
        >>> from paradex_py import Paradex
        >>> from paradex_py.environment import Environment
        >>> paradex = Paradex(env=Environment.TESTNET)
        >>> paradex.ws_client.connect()
    """

    def __init__(
        self,
        env: Environment,
        logger: Optional[logging.Logger] = None,
    ):
        self.env = env
        self.api_url = f"wss://ws.api.{self.env}.paradex.trade/v1"
        self.logger = logger or logging.getLogger(__name__)
        self.ws: Optional[websockets.WebSocketClientProtocol] = None

    async def __aexit__(self):
        await self.close_connection()

    def init_account(self, account: ParadexAccount) -> None:
        self.account = account

    async def connect(self) -> None:
        try:
            extra_headers = {}
            if self.account:
                extra_headers.update({"Authorization": f"Bearer {self.account.jwt_token}"})
            self.ws = await websockets.connect(
                self.api_url,
                extra_headers=extra_headers,
            )
            self.logger.info(f"Paradex_WS: Connected to {self.api_url}")
            if self.account:
                await self.send_auth_id(self.ws, self.account.jwt_token)
                self.logger.info(f"Paradex_WS: Authenticated to {self.api_url}")
        except (
            websockets.exceptions.ConnectionClosedOK,
            websockets.exceptions.ConnectionClosed,
        ) as e:
            self.logger.info(f"Paradex_WS connection already closed:{e}")
            self.ws = None
        except Exception as e:
            # Reduced from error to warning to avoid flood on Sentry
            self.logger.warning(f"error:{e} traceback:{traceback.format_exc()}")
            self.ws = None

    async def close_connection(self):
        try:
            if self.ws:
                self.logger.info("Paradex_WS Closing connection...")
                await self.ws.close()
                self.logger.info("Connection closed")
            else:
                self.logger.info("Paradex_WS: No connection to close")
        except Exception:
            self.logger.exception(f"Error thrown when closing connection {traceback.format_exc()}")

    async def reconnect(self):
        self.logger.info("Paradex_WS to reconnect websocket...")
        await self.close_connection()
        await self.connect()

    async def send_auth_id(
        self,
        websocket: websockets.WebSocketClientProtocol,
        paradex_jwt: str,
    ) -> None:
        """
        Sends an authentication message to the Paradex WebSocket.
        """
        await websocket.send(
            json.dumps(
                {
                    "id": int(time.time() * 1_000_000),
                    "jsonrpc": "2.0",
                    "method": "auth",
                    "params": {"bearer": paradex_jwt},
                }
            )
        )

    async def read_messages(self, read_timeout=0.1, backoff=0.1):
        while True:
            try:
                response = await asyncio.wait_for(self.ws.recv(), timeout=read_timeout)
                message = json.loads(response)
                if "id" in message:
                    channel_subscribed: Optional[str] = message.get("result", {}).get("channel")
                    if channel_subscribed:
                        self.logger.info(f"Paradex_WS: subscribed to channel:{channel_subscribed}")
                yield message
            except (
                websockets.exceptions.ConnectionClosedError,
                websockets.exceptions.ConnectionClosedOK,
            ):
                self.logger.exception(f"Paradex_WS: connection closed {traceback.format_exc()}")
                await self.reconnect()
            except asyncio.TimeoutError:
                await asyncio.sleep(backoff)
            except Exception:
                self.logger.exception(f"Paradex_WS: connection failed {traceback.format_exc()}")
                await asyncio.sleep(1)

    async def _send(self, message: str):
        try:
            if self.ws:
                await self.ws.send(message)
        except websockets.exceptions.ConnectionClosedError as e:
            self.logger.info(f"Paradex_WS: Restarted connection {e}")
            await self.reconnect()
            if self.ws:
                await self.ws.send(message)
        except Exception:
            self.logger.exception(f"Paradex_WS: send failed {traceback.format_exc()}")
            await self.reconnect()

    async def subscribe(
        self,
        channel: ParadexWebsocketChannel,
        market: Optional[str] = None,
    ) -> None:
        channel_name = channel.value.format(market=market)
        self.logger.info(f"Paradex_WS: subscribe {channel}/{market} name:{channel_name}")
        await self._send(
            json.dumps(
                {
                    "id": int(time.time() * 1_000_000),
                    "jsonrpc": "2.0",
                    "method": "subscribe",
                    "params": {"channel": channel_name},
                }
            )
        )
